/* qemu2qlay */

#include <stdio.h>
#include <string.h>
#include <assert.h>

#define MAX_SECTORS 255

typedef struct {
  unsigned char preamble[12];   // 10*0, 2*ff
  unsigned char ff;  
  unsigned char snum;  
  char name[10];
  unsigned char rnd[2];
  unsigned short csum;
} __attribute__((packed)) hdr_t;

typedef struct {
  unsigned char bh_preamble[12];   // 10*0, 2*ff
  unsigned char file;
  unsigned char block;
  unsigned short bh_csum;

  unsigned char data_preamble[8];   // 6*0, 2*ff
  unsigned char data[512];
  unsigned short data_csum;

  unsigned char extra_byte[120];
} __attribute__((packed)) sector_t;

typedef struct {
  hdr_t hdr;
  sector_t sec;
} mdv_entry_t;

// buffer for mdv file
mdv_entry_t buffer[MAX_SECTORS];

unsigned short sum(unsigned char *p, int len) {
  unsigned short v = 0x0f0f;
  while(len--)
    v += *p++;

  return v;
}

typedef struct {
  char magic[8];   // Mdv*Dump
  int a;
  int b;
  short c;
  char sectors0;
  char sectors1;
  char stuff[26];
} __attribute__((packed)) qlem_hdr_t;

int main(int argc, char **argv) {
  FILE *in, *out;

  if(argc != 3) {
    printf("Usage: qemu2qlay <mdv_in> <mdv_out>\n");
    return 0;
  }

  // clean output buffer
  int i;
  for(i=0;i<MAX_SECTORS;i++) {
    memset(&buffer[i], 0, sizeof(mdv_entry_t));
    
    // setup header
    buffer[i].hdr.preamble[10] = 0xff;
    buffer[i].hdr.preamble[11] = 0xff;
    buffer[i].hdr.csum = 0x0f0f;

    // setup block header
    buffer[i].sec.bh_preamble[10] = 0xff;
    buffer[i].sec.bh_preamble[11] = 0xff;
    buffer[i].sec.bh_csum = 0x0f0f;

    // data
    buffer[i].sec.data_preamble[6] = 0xff;
    buffer[i].sec.data_preamble[7] = 0xff;
    buffer[i].sec.data_csum = 0x0f0f;
    memset(buffer[i].sec.extra_byte, 0x5a, 
	   sizeof(buffer[i].sec.extra_byte));
  }


  if(!(in = fopen(argv[1], "rb"))) {
    fprintf(stderr, "error opening input file\n");
    return -1;
  }

  // do some sanity checks on input file

  // check if file size makes sense
  int size = 0;
  fseek(in, 0, SEEK_END);
  size = ftell(in);
  fseek(in, 0, SEEK_SET);

  if((size-sizeof(qlem_hdr_t))%(512+16) != 0) {
    fprintf(stderr, "input file size error\n");
    return -1;
  }

  qlem_hdr_t hdr;
  
  if(fread(&hdr, sizeof(qlem_hdr_t), 1, in) != 1) {
    fprintf(stderr, "error reading header\n");
    return -1;
  }

  if(strcmp(hdr.magic, "Mdv*Dump") != 0) {
    fprintf(stderr, "input file is not a qlem mdv image\n");
    return -1;
  }

  int sectors = (size-sizeof(qlem_hdr_t))/(512+16);
  printf("Converting %d sectors ...\n", sectors );  

  for(i=0;i<sectors;i++) {
    struct {
      unsigned char sec_hdr[14];
      unsigned char block_hdr[2];
      unsigned char data[512];
    } __attribute__((packed)) qlem_sector;
    
    if(fread(&qlem_sector, sizeof(qlem_sector), 1, in) != 1) {
      fprintf(stderr, "error reading sector %d\n", i);
      return -1;
    }

    // copy sector header
    memcpy(&buffer[i].hdr.ff, qlem_sector.sec_hdr, sizeof(qlem_sector.sec_hdr));
    // set checksum
    buffer[i].hdr.csum = sum(&buffer[i].hdr.ff, sizeof(qlem_sector.sec_hdr));

    // copy block header
    memcpy(&buffer[i].sec.file, qlem_sector.block_hdr, sizeof(qlem_sector.block_hdr));
    // set checksum
    buffer[i].sec.bh_csum = sum(&buffer[i].sec.file, sizeof(qlem_sector.block_hdr));
    
    // copy data
    memcpy(&buffer[i].sec.data, qlem_sector.data, sizeof(qlem_sector.data));
    // set checksum
    buffer[i].sec.data_csum = sum((char*)&buffer[i].sec.data, sizeof(qlem_sector.data));
  }

  fclose(in);

  out = fopen(argv[2], "wb");
  if(!out) {
    fprintf(stderr, "error opening output file\n");
    return -1;
  }

  if(fwrite(buffer, sizeof(mdv_entry_t), MAX_SECTORS, out) != MAX_SECTORS) {
    fprintf(stderr, "error writing qlay image\n");
    return -1;
  }

  fclose(out);

  return 0;
}
